"""
Data structures used in translation of the Doconce format to other formats.
Some functions used in translation modules are also included here.
"""
import re

def indent_lines(text, format, indentation=' '*8):
    if format == 'epytext':
        # some formats (Epytext) give wrong behavior if there is a \n
        # in code or tex blocks (\nabla is one example), hence we
        # comment out the block in those cases: (\b also gives problems...
        # it seems that Epytext is not capable of doing raw strings here)
        if re.search(r'\\n', text):
            comment_out = True
            text = """\
            
            NOTE: A verbatim block has been removed because
                  it causes problems for Epytext.
"""
            return text
        
    # indent X chars (choose X=6 for sufficient indent in lists)
    text = '\n'.join([indentation + line for line in text.splitlines()])
    return text

def remove_code_and_tex(filestr):
    # store code and tex blocks in lists and substitute these blocks
    # by #!!CODE_BLOCK and #!!TEX_BLOCK ("arguments" to !bc must be
    # copied after #!!CODE_BLOCK).
    # later we replace #!!CODE_BLOCK by !bc and the code block again
    # (similarly for the tex block).
    
    # (recall that !bc can be followed by extra information that we must keep:)
    code = re.compile(r'^!bc(.*?)\n(.*?)!ec\n', re.DOTALL|re.MULTILINE)
    code_blocks = [c for opt, c in code.findall(filestr)]
    tex = re.compile(r'^!bt\n(.*?)!et\n', re.DOTALL|re.MULTILINE)
    tex_blocks = tex.findall(filestr)
    # remove:
    filestr = code.sub('#!!CODE_BLOCK \g<1>\n', filestr)
    filestr = tex.sub('#!!TEX_BLOCK\n', filestr)
    return filestr, code_blocks, tex_blocks
        
def insert_code_and_tex(filestr, code_blocks, tex_blocks, format):
    lines = filestr.splitlines()
    for code in code_blocks:
        # this construction does not handle newlines in regex and
        # similar verbatim environments properly:
        #filestr = re.sub(r'#!!CODE_BLOCK (.*?)\n',
        #                 '!bc\g<1>\n%s!ec\n' % code, filestr, 1)
        # use string.replace instead
        
        for i in range(len(lines)):
            if '#!!CODE_BLOCK' in lines[i]:
                # use re.sub for the heading when we need a group:
                lines[i] = re.sub('#!!CODE_BLOCK(.*)',
                                  '!bc\g<1>\n!XX&XX', lines[i])
                # use string.replace to deal correctly with \n:
                lines[i] = lines[i].replace('!XX&XX', '%s!ec' % code)
                break
    for tex in tex_blocks:
        # also here problems with this: (\nabla becomes \n (newline) and abla)
        #filestr = re.sub(r'#!!TEX_BLOCK', '!bt\n%s!et' % tex, filestr, 1)
        for i in range(len(lines)):
            if '#!!TEX_BLOCK' in lines[i]:
                lines[i] = lines[i].replace('#!!TEX_BLOCK',
                                            '!bt\n%s!et' % tex)
                break
    filestr = '\n'.join(lines)
    return filestr

BLANKLINE = {}
FILENAME_EXTENSION = {}
LIST = {}
CODE = {}
# listing of function parameters/arguments and return values:
ARGLIST = {}
DEFAULT_ARGLIST = {
    'parameter': 'argument',
    'keyword': 'keyword argument',
    'return': 'return value(s)',
    'instance variable': 'instance variable',
    'class variable': 'class variable',
    'module variable': 'module variable',
    }
TABLE = {}
FIGURE_EXT = {}
INTRO = {}
OUTRO = {}

        
# regular expressions for inline tags:
inline_tag_begin = r"""(?P<begin>(^|[(\s]))"""
inline_tag_end = r"""(?P<end>($|[.,?!;:)\s]))"""
INLINE_TAGS = {
    # math: text inside $ signs, as in $a = b$, with space before the
    # first $ and space, comma, period, colon, semicolon, or question
    # mark after the enclosing $.
    'math':
    r'%s\$(?P<subst>[^ `][^$`]*)\$%s' % \
    (inline_tag_begin, inline_tag_end),

    # $latex text$|$pure text alternative$
    'math2':
    r'%s\$(?P<latexmath>[^ `][^$`]*)\$\|\$(?P<puretext>[^ `][^$`]*)\$%s' % \
    (inline_tag_begin, inline_tag_end),

    # *emphasized words*
    'emphasize':
    r'%s\*(?P<subst>[^ `][^*`]*)\*%s' % \
    (inline_tag_begin, inline_tag_end),

    # `verbatim inline text is enclosed in back quotes`
    'verbatim':
    r'%s`(?P<subst>[^ ][^`]*)`%s' % \
    (inline_tag_begin, inline_tag_end),

    # _underscore before and after signifies bold_
    'bold':
    r'%s_(?P<subst>[^ `][^_`]*)_%s' % \
    (inline_tag_begin, inline_tag_end),

    # label{labelname}
    'label':
    r'label\{(?P<subst>[^}]+)\}',

    # ref{labelname}
    'reference':
    r'ref\{(?P<subst>[^}]+)\}',

    # http://some.where.org/mypage<link text>
    'linkURL':
    r'%s(?P<url>http://[^<]+)<(?P<link>[^>]+)>%s' % \
    (inline_tag_begin, inline_tag_end),

    'plainURL': # problems with this one
    r'%s(?P<url>http://[^,?!;:)\s<]+)(?P<end>[,?!;:)\s])' % \
    (inline_tag_begin),

    # _______Seven Underscores Before and After Section Title_______
    'section':
    #r'^_{7}(?P<subst>[^ ].*)_{7}\s*$',
    r'^\s*_{7}(?P<subst>[^ ].*?)_+\s*$',

    # _____Five Underscores Before and After Subsection Title_____
    'subsection':
    #r'^_{5}(?P<subst>[^ ].*)_{5}\s*$',
    r'^\s*_{5}(?P<subst>[^ ].*?)_+\s*$',

    # ___Three Underscores Before and After Subsubsection Title___
    'subsubsection':
    #r'^_{3}(?P<subst>[^ ].*)_{3}\s*$',
    r'^\s*_{3}(?P<subst>[^ ].*?)_+\s*$',

    # __Two Underscores Before and After Inline Paragraph Title.__
    'paragraph':
    r'(?P<begin>^)_{2}(?P<subst>.+?)_+\s+',

    # TITLE: My Document Title
    'title':
    r'^TITLE:\s*(?P<subst>.+)\s*$',

    # AUTHOR: Some Name
    'author':
    r'^AUTHOR:\s*(?P<subst>.+)\s*$',

    # DATE: Jan 27, 2010
    'date':
    r'^DATE:\s*(?P<subst>.+)\s*$',

    # FIGURE:[filename, options] some caption text label{some:label}
    'figure':
    r'^FIGURE:\[(?P<filename>[^,\]]+),?(?P<options>[^\]]*)\]\s*(?P<caption>.+)$',
    }
INLINE_TAGS_SUBST = {}

# frequent syntax errors that we can test for: (not yet used)
heading_error = (r'(^ _+[^_]*|^_+[^ _]*$)',
                 'Initial spaces or missing underscore(s) at the end')
INLINE_TAGS_BUGS = {
    # look for space after first special character ($ ` _ etc)
    'math': (r'%s(?P<subst>\$ [^$]*\$)%s' % (inline_tag_begin, inline_tag_end),
             'Space after first $ in inline math expressions'),
    'emphasize': None, # *item with *emph* word is legal (no error)
    'verbatim': (r'%s(?P<subst>` [^`]*`)%s' % \
                 (inline_tag_begin, inline_tag_end),
                 'Space after first ` in inline verbatim expressions'),
    'bold': (r'%s(?P<subst>_ [^_]*_)%s' % \
             (inline_tag_begin, inline_tag_end),
             'Space after first _ in inline boldface expressions'),
    # heading errors: spaces first or missing _ at the end
    'section': heading_error,
    'subsection': heading_error,
    'subsubsection': heading_error,
    'paragraph': heading_error,
    }

LIST_SYMBOL = {'*': 'itemize', 'o': 'enumerate', '-': 'description'}
