#!/usr/bin/env python
'''




What Is Doconce?
================

Doconce may be two things:

 1. Doconce is a working strategy for documenting software in a single
    place and avoid duplication of information. The idea is to write
    once and include that text in many different documents (manuals,
    doc strings, tutorials, books, etc.) in many different formats.

 2. Doconce is a simple and minimally tagged markup language for producing
    nicely formatted text for software documentation, reports, manuals,
    thesis, and books.  You can write almost plain ASCII text according to
    some simple formatting rules, and a Python script 
    doconce2format can automatically transform this text 
    to HTML, LaTeX, reStructuredText, Epytext, PDF, etc, and 
    even plain text (with non-obvious formatting/tags removed for clear 
    reading).

Point 1 may be of interest even if you adopt a different markup language
than Doconce (e.g., reStructuredText). You can jump to
the section "The Doconce Documentation Strategy" to see a recipe
for how to perform step 1 and 2 above.


Motivation: Problems with Documenting Software
----------------------------------------------

*Duplicated Information.* It is common to write some software
documentation in the code (doc strings in Python, doxygen in C++,
javadoc in Java) while similar documentation is often also included in
a LaTeX or HTML manual or tutorial. Although the various types of
documentation may start out to be the same, different physical files
must be used since very different tagging is required for different
output formats. Over time the duplicated information starts to
diverge. Severe problems with such unsynchronized documentation was
the motivation for developing the Doconce concept and tool.

*Different Tagging for Different Formats.* A problem with doc strings
(in Python) is that they benefit greatly from some tagging, Epytext or
reStructuredText, when transformed to Epydoc HTML/PDF manuals. However, 
such tagging looks annoying in Pydoc - there we should have more
minimal (or no) tagging since Pydoc only presents the doc strings
verbatim (students and newbies are in particular annoyed by any
unfamiliar tagging of ASCII text). Manuals/tutorials in HTML and 
LaTeX need much more tagging.

*Solution.* Accurate information is crucial and can only be maintained
in a *single physical* file, which must be filtered to suitable
formats and included in various documents (LaTeX manuals/tutorials,
Pydoc/Epydoc/HappyDoc reference manuals).

*A Common Format.* There is no existing format and associated
conversion tools that allow a "singleton" documentation file to be
filtered to LaTeX, HTML, XML, PDF, Epydoc, HappyDoc, Pydoc, and plain
untagged text. As we are involved with mathematical software, the
LaTeX manuals should have nicely typeset mathematics, while Pydoc,
Epydoc, and HappyDoc must show LaTeX math in verbatim mode.
Unfortunately, Epytext is annoyed by even very simple LaTeX math (also
in verbatim environments). To summarize, we need

 1. A minimally tagged markup language with full support for 
    for mathematics and verbatim computer code.

 2. Filters for producing highly tagged formats (LaTeX, HTML, XML),
    medium tagged formats (reStructuredText, Epytext), and plain
    text with completely invivisble tagging. 

 3. Tools for inserting appropriately filtered versions of a "singleton"
    documentation file in other documents (doc strings, manuals, tutorials).

The answer to these points is the Doconce markup language, its associated
tools, and the preprocess.py C-like preprocessor (by Mick Trent, for lots
of languages). Then we can *write once, include everywhere!*
...And what we write is close to plain ASCII text.

The resulting tools and methodology were inspired by Python code documentation,
but can be applied to most other languages (support for Javadoc and Doxygen
is not yet included, but should be very straightforward).


The Doconce Documentation Strategy
----------------------------------

   * Write software documentation, both tutorials and manuals, in
     the Doconce format. Use many files and never duplicate information!!

   * Use #include statements in source code (especially in doc strings)
     and in LaTeX documents for including documentation files.
     These documentation files must be filtered to an appropriate format
     by doconce2format before being included. In a Python context, this
     means plain text for computer source code (and Pydoc), Epytext or 
     reStructuredText for Epydoc, LaTeX for LaTeX manuals, or
     reStructuredText for XML, Docbook, OpenOffice, RTF (the latter three
     will appear as Docutils rst2*.py scripts in the near future).

   * Run the preprocessor preprocess (by Mick Trent) on the files to 
     produce native files for pure computer code and for various other 
     documents.

Consider an example involving a Python module in a basename.p.py file.
The .p.py extension identifies this as a file that has to be
preprocessed) by the preprocess program. 
In a doc string in basename.p.py we do a preprocessor include, say::

        #  #include "doc/doc1.dst.txt

which is a file filtered to a specific format (typically plain text or
Epytext) from an original "singleton" documentation file named
doc/doc1.do.txt. The .do.txt extension signifies a file in Doconce
format (because of the end extension .txt, editors and other
applications treat such as plain text files, which is convenient.)

For making an Epydoc manual, the doc/doc1.do.txt file is filtered to
doc/doc1.epytext and renamed to doc/doc1.dst.txt (.dst for Doc
STring, .txt for the preprocessor to accept this as a known file
type...).  Then we run the preprocessor on the basename.p.py file
and create a real Python file basename.py extensions. Finally, we
run Epydoc on this file:

The next step might be to produce pure Python source code. For this
purpose we filter doc/doc1.do.txt to plain text format (doc/doc1.txt)
and renamed to doc/doc1.dst.txt. The preprocessor transforms the
basename.p.py file to a basename.py for import and execution. The
doc strings are now in plain text and well suited for Pydoc.
All these steps are automated by the insertdocstr.py script.
Here are the Unix commands::

        # make Epydoc API manual of basename module:
        cd doc
        doconce2format epytext doc1.do.txt
        mv doc1.epytext doc1.dst.txt
        cd ..
        preprocess basename.p.py > basename.py
        epydoc basename
        
        # make ordinary Python module files with doc strings:
        cd doc
        doconce2format plain doc1.do.txt
        mv doc1.txt doc1.dst.txt
        cd ..
        preprocess basename.p.py > basename.py
        
        # can automate inserting doc strings in all .p.py files:
        insertdocstr.py plain .
        # (runs through all .do.txt files and filters them to plain format and
        # renames to .dst.txt extension, then the script runs through all 
        # .p.py files and runs the preprocessor, which includes the .dst.txt
        # files)


For a LaTeX manual we filter doc/doc1.do.txt to LaTeX format, resulting
in doc/doc1.tex, maybe interpret this text by script, and make it
ready to be included in a larger LaTeX document. Similar things can
be done in HTML format. Or one can go to reStructuredText and from
there to XML, DocBook, LaTeX, HTML.


A Demo
------

The current text is generated from a Doconce format stored in the::

        lib/doconce/doc/doconce.do.txt

file in the Doconce source code tree.
The file make.sh in the same directory shows how to run doconce2format
on the doconce.do.txt file to obtain documents in various formats.  To
go from the LaTeX format to PDF, see latex.sh.
Running this demo (make.sh and latex.sh) and studying the various
generated files and comparing them with the original doconce.do.txt
file, gives a quick introduction to how Doconce is used in a real case.


Why Yet Another Minimally Tagged Markup Language?
-------------------------------------------------

The Python community has already produced two successful, ASCII-based
markup languages with minimal tagging: reStructuredText
(part of Docutils (http://www.docutils.org)) and Epytext (part of 
Epydoc (http://epydoc.sourceforge.net/)).
(Other minimum-tagged markup 
languages do also exist (http://www.pault.com/xmlalternatives.html)). So why is Doconce needed?

  Simplicity: 
    Doconce supports only a handful of formatting constructions and
    is therefore simpler than Epytext and reStructuredText.
    Consequently, the latter two are more flexible. However,
    their parsing modules are also much more complex, and it is
    an order of magnitude simpler to add support for a new format
    in Doconce than in other markup languages. 

    A small set of formatting constructions also helps to keep the
    formatting functional even in plain ASCII output. 

    For documenting software it might be a good idea to keep the
    formatting simple when the same text is used both in the code,
    in ASCII man pages, in HTML, and in LaTeX. Even more
    important, the simple Doconce format is less cluttered and easier 
    to read than more flexible/advanced ASCII-based markup languages.

  Flexibility: 
    Epydoc produces complete module/package documentation.
    Doconce can work with small pieces of documentation, e.g.,
    module doc strings to be filtered to and used by Epydoc, or
    included in a LaTeX thesis or manual.

  Math: 
    Doconce supports inline mathematical expressions and blocks
    of mathematical LaTeX code. (Epytext is easily fooled by TeX
    code with the backslashes and give lots of error messages
    even if the code block is to be typeset in varbatim mode.) 

  More formats: 
    Doconce can currently be translated to HTML, LaTeX, reStructuredText,
    StructuredText, Epytext, and plain (almost untagged) ASCII.

    Epytext and reStructuredText can only be transformed to
    HTML and PDF. Doconce can also be transformed to more straight
    ASCII text, for example online manuals intended to be read
    in terminal windows. When Doconce is transformed to reStructuredText
    the text can further be transformed to HTML, LaTeX, and XML.
    reStructuredText support for Docbook, RTF, and OpenOffice is
    under development. Doconce can also be transformed to Epytext for
    Epydoc formatting of modules, or to StructuredText used by
    HappyDoc for Python modules.

    A page of code, containing definitions of how emphasized text, 
    verbatim text, lists, etc. are rendered, is all you need to 
    implement a new format.
    (Everything except lists are parsed by regular expressions in Doconce.)




The Doconce Markup Language
===========================

The Doconce format introduces four constructs to markup text:
lists, special lines, inline tags, and environments.

Lists
-----

An unordered bullet list makes use of the * as bullet sign
and is indented as follows::

           * item 1
        
           * item 2
        
             * subitem 1, if there are more
               lines, each line must
               be intended as shown here
        
             * subitem 2,
               also spans two lines
        
           * item 3


This list gets typeset as

   * item 1

   * item 2

     * subitem 1, if there are more
       lines, each line must
       be intended as shown here

     * subitem 2,
       also spans two lines

   * item 3

In an ordered list, each item starts with an o (as the first letter 
in "ordered")::

           o item 1
        
           o item 2
        
             * subitem 1
        
             * subitem 2
        
           o item 3


resulting in

  1. item 1

  2. item 2

     * subitem 1

     * subitem 2

  3. item 3

Ordered lists cannot have an ordered sublist, i.e., the ordering 
applies to the outer list only.

In a description list, each item is recognized by a dash followed
by a keyword followed by a colon::

           - keyword1: explanation of keyword1
        
           - keyword2: explanation
             of keyword2 (remember to indent properly
             if there are multiple lines)


The result becomes

   keyword1: 
     explanation of keyword1

   keyword2: 
     explanation
     of keyword2 (remember to indent properly
     if there are multiple lines)

Special Lines
-------------

The Doconce markup language has a concept called *special lines*.
Such lines starts with a markup at the very beginning of the
line and are used to mark document title, authors, date,
sections, subsections, paragraphs., figures, etc.

Lines starting with TITLE:, AUTHOR:, and DATE: are optional and used
to identify a title of the document, the authors, and the date. The
title is a treated as the rest of the line, so is the date, but the
author text consists of the name and associated institutions,
separated by semicolons. Multiple authors require multiple AUTHOR:
lines. All information associated with TITLE: and AUTHOR: keywords
must appear on a single line.  Here is an example::

        TITLE: On The Ultimate Markup Language: Doconce
        AUTHOR: H. P. Langtangen, Simula Research Laboratory, Univ. of Oslo
        AUTHOR: Kaare Dump; Segfault Inc., Cyberspace
        DATE: November 9, 2006


Headlines are recognized by being surrounded by underscores before and
after the text of the headline. Different section levels are recognized 
by the associated number of underscores:
   * 7 underscores for sections
   * 5 underscores for subsections
   * 3 underscroes for subsubsections
   * 2 underscores for paragraphs (paragraph heading will be inlined)

Here are some examples::

        _______A Section Heading_______
        
        The running text goes here.
        
              _____A Subsection Heading_____
        The running text goes here.
        
                  ___A Subsubsection Heading___
        
        The running text goes here.
        
        __A Paragraph.__ The running text goes here.


The result for the present format looks like this:

A Section Heading
=================

The running text goes here.

A Subsection Heading
--------------------
The running text goes here.

A Subsubsection Heading
~~~~~~~~~~~~~~~~~~~~~~~

The running text goes here.

*A Paragraph.* The running text goes here.


Figures are recognized by the special line syntax::

        FIGURE:[filename, height=xxx width=yyy scale=zzz] caption

The height, width, and scale keywords (and others) can be included
if desired and may have effect for some formats. Note the comma
between the sespecifications and that there should be no space
around the = sign.

Note also that, like for TITLE: and AUTHOR: lines, all information
related to a figure line must be written on the same line. Introducing
newlines in a long caption will destroy the formatting (only the
part of the caption appearing on the same line as FIGURE: will be
included in the formatted caption).

The filename extension may not be compatible with the chosen output format.
For example, a filename mypic.eps is fine for LaTeX output but not for
HTML. In such cases, the Doconce translator will convert the file to
a suitable format (say mypic.png for HTML output).

FIGURE:[figs/dinoimpact.gif, width=400] It can't get worse than this.... fig:impact



Another type of special lines starts with @@@CODE and enables copying
of computer code directly into a verbatim environment, see 
the section "Blocks of Verbatim Computer Code" below.


Inline Tagging
--------------

Doconce supports tags for *emphasized phrases*,
_boldface phrases_, and verbatim text (also called type writer text, for
inline code) plus LaTeX/TeX inline mathematics, such as a = \sin(x).

   * Emphasized text is typeset inside a pair of asterix, and there should
     be no spaces between an asterix and the emphasized text, as in::

        *emphasized words*


   * Boldface font is recognized by an underscore instead of an asterix::

        _several words in boldface_ followed by *ephasized text*.


     The line above gets typeset as
     _several words in boldface_ followed by *ephasized text*.

   * Verbatim text, typically used for short inline code,
     is typeset between backquotes::

        `call myroutine(a, b)` looks like a Fortran call
        while `void myfunc(double *a, double *b)` must be C.

     The typesetting result looks like this:
     call myroutine(a, b) looks like a Fortran call
     while void myfunc(double *a, double *b) must be C.

     Watch out for mixing backquotes and asterix (i.e.,
     verbatim and emphasized code): the Doconce
     interpreter is not very smart so inline C code can soon lead
     to problems in the final format. Go back to the Doconce
     source and modify it so the format to which you want to go becomes
     correct (sometimes a trial and error process - sticking to
     very simple formatting usually avoids such problems).

   * URLs with links are typeset as::

        some URL http://my.place.in.space/src<MyPlace>

     which appears as some URL MyPlace (http://my.place.in.space/src).

   * References and labels are supported. The syntax is simple::

        label{section:verbatim}   # defines a label
        For more information we refer to Section ref{section:verbatim}.


     LaTeX and HTML support such labels and references, and the label
     name are invisible in the resulting document. For other formats,
     especially plain text, Epytext and similar, the names inside the
     label and ref specifications are just printed in the document
     so the names should be carefully chosen to help the reader understand
     what is meant by the label or reference. One suggestion is to
     apply type:name as illustrated above (section:verbatim),
     where type is the type of document element (section, equation, 
     figure, etc.) and name is its logical name. When transforming
     Doconce to LaTeX, figure captions automatically get a label with
     a name identical to the basename of the figure file.

     By the way, here is an example on referencing Figure fig:impact
     (see the figure caption for an explicit label).


   * Inline mathematics is written as in LaTeX, i.e., inside dollar
     signs.  Most formats leave this syntax as it is (including to
     dollar signs), hence nice math formatting is only obtained in
     LaTeX (Epytext has some inline math support that is utilized).
     However, mathematical expressions in LaTeX syntax often contains
     special formatting commands (such as user-defined newcommands),
     which appear annoying in plain text. Doconce therefore supports
     an extended inline math syntax where the writer can provide an
     alternative syntax suited for formats close to plain ASCII::

        Here is an example on a linear system 
        ${\bf A}{\bf x} = {\bf b}$|$Ax=b$, 
        where $\bf A$|$A$ is an $n\times n$|$nxn$ matrix, and 
        $\bf x$|$x$ and $\bf b$|$b$ are vectors of length $n$|$n$.


     The above text is typeset as "Here is an example on a linear
     system Ax=b, where A is an
     nxn matrix, and x and b are
     vectors of length n."


Blocks of Verbatim Computer Code
--------------------------------

Blocks of computer code, to be typeset verbatim, must appear inside
a "begin code" !bc keyword and an "end code" !ec keyword. Both
keywords must be on a single line and start at the beginning of the line
to be recognized. It is important that the !bc tag starts at the
beginning of the line. The rest of the line may contain instructions
to tailor the formatting (this text is only copied to the output file
by Doconce). The !ec tag must start in column 1 and be followed by
a newline.

Here is a verbatim code block::

        # regular expressions for inline tags:
        inline_tag_begin = r'(?P<begin>(^|\s+))'
        inline_tag_end = r'(?P<end>[.,?!;:)\s])'
        INLINE_TAGS = {
            'emphasize':
            r'%s\*(?P<subst>[^ `][^*`]*)\*%s' % \
            (inline_tag_begin, inline_tag_end),
            'verbatim':
            r'%s`(?P<subst>[^ ][^`]*)`%s' % \
            (inline_tag_begin, inline_tag_end),
            'bold':
            r'%s_(?P<subst>[^ `][^_`]*)_%s' % \
            (inline_tag_begin, inline_tag_end),
        }


Computer code can be copied directly from a file, if desired. The syntax
is then::

        @@@CODE myfile.f
        @@@CODE myfile.f fromto:subroutine\s+test@^C\s{5}END1

The first line implies that all lines in the file myfile.f are copied
into a verbatim block. The second line has a `fromto:' directive, which
implies copying code between two lines in the code. Two regular
expressions, separated by the @ sign, define the "from" and "to" lines.
The "from" line is included in the verbatim block, while the "to" line
is not. In the example above, we copy code from the line matching
subroutine test (with as many blanks as desired between the two words)
and the line matching C      END1 (an end comment - this line is not
included in the verbatim block).

Let us copy a whole file (the first line above)::

        C     a comment
        
              subroutine    test()
              integer i
              real*8 r
              r = 0
              do i = 1, i
                 r = r + i
              end do
              return
        C     END1
        
              program testme
              call test()
              return
        
        



Let us then copy just a piece in the middle as indicated by the fromto:
directive above::

              subroutine    test()
              integer i
              real*8 r
              r = 0
              do i = 1, i
                 r = r + i
              end do
              return
        




LaTeX Blocks of Mathematical Text
---------------------------------

Blocks of mathematical text are like computer code blocks, but
the opening tag is !bt (begin TeX) and the closing tag is
!et. It is important that !bt and !et appear on the beginning of the
line followed by a newline. 

Here is the result of a !bt - !et block::

        \begin{eqnarray}
        {\partial u\over\partial t} &=& \nabla^2 u + f,\label{myeq1}\\
        {\partial v\over\partial t} &=& \nabla\cdot(q(u)\nabla v) + g
        \end{eqnarray}


This text looks ugly in all Doconce supported formats, except from
LaTeX.  If HTML is desired, the best is to filter the Doconce text
first to LaTeX and then use the widely available tex4ht tool to
convert the dvi file to HTML. For other textual formats, it is best to
avoid blocks of mathematics and instead use inline mathematics where
it is possible to write expressions both in native LaTeX format (so it
looks professional in LaTeX) and in a pure text format (so it at least
looks okay).


Header and Footer
-----------------

Some formats use a header and footer in the document. LaTeX and
HTML are two examples of such formats. When the document is to be
included in another document (which is often the case with
Doconce-based documents), the header and footer are not wanted, while
these are needed (at least in a LaTeX context) if the document is
stand-alone. We have introduce the convention that if TITLE: or
#TITLE: is found at the beginning of the line (i.e., the document
has, or has an intention have, a title), the header and footer
are included, otherwise not.


Basic Parsing Ideas
-------------------

 1. Parts of files with computer code are first copied into verbatim
    blocks.

 2. All verbatim and TeX blocks are first removed and stored elsewhere
    to ensure that formatting rules are not applied to these blocks.

 3. The text is examined line by line for typesetting of lists.
    List parsing needs some awareness of the context.
    Each line is interpreted by a regular expression::

        (?P<indent> *(?P<listtype>[*o-] )? *)(?P<keyword>[^:]+?:)?(?P<text>.*)\s?


    That is, a possible indent (which we measure), an optional list
    item identifier, optional space, optional words ended by colon,
    and optional text. All lines are of this form. However, some
    ordinary (non-list) lines may contain a colon, and then the keyword
    and text group must be added to get the line contents. Otherwise,
    the text group will be the line.

 4. When lists are typeset, the text is examined for sections, paragraphs,
    title, author, date, plus all the inline tags for emphasized, boldface,
    and verbatim text. Plain subsitutions based on regular expressions
    are used for this purpose.

 5. The final step is to insert the code and TeX blocks again (these should
    be untouched and are therefore left out of the previous parsing).


A Glimpse of How to Write a New Translator
------------------------------------------

This is the HTML-specific part of the
source code of the HTML translator::

        FILENAME_EXTENSION['HTML'] = '.html'  # output file extension
        BLANKLINE['HTML'] = '<p>\n'           # blank input line => new paragraph
        INLINE_TAGS_SUBST['HTML'] = {         # from inline tags to HTML tags
            # keep math as is:
            'math': None,  # indicates no substitution
            'emphasize':     r'\g<begin><em>\g<subst></em>\g<end>',
            'bold':          r'\g<begin><b>\g<subst></b>\g<end>',
            'verbatim':      r'\g<begin><tt>\g<subst></tt>\g<end>',
            'URL':           r'\g<begin><a href="\g<url>">\g<link></a>',
            'section':       r'<h1>\g<subst></h1>',
            'subsection':    r'<h3>\g<subst></h3>',
            'subsubsection': r'<h5>\g<subst></h5>',
            'paragraph':     r'<b>\g<subst></b>. ',
            'title':         r'<title>\g<subst></title>\n<center><h1>\g<subst></h1></center>',
            'date':          r'<center><h3>\g<subst></h3></center>',
            'author':        r'<center><h3>\g<subst></h3></center>',
            }
        
        # how to replace code and LaTeX blocks by HTML (<pre>) environment:
        def HTML_code(filestr):
            c = re.compile(r'^!bc(.*?)\n', re.MULTILINE)
            filestr = c.sub(r'<!-- BEGIN VERBATIM BLOCK \g<1>-->\n<pre>\n', filestr)
            filestr = re.sub(r'!ec\n',
                             r'</pre>\n<! -- END VERBATIM BLOCK -->\n', filestr)
            c = re.compile(r'^!bt\n', re.MULTILINE)
            filestr = c.sub(r'<pre>\n', filestr)
            filestr = re.sub(r'!et\n', r'</pre>\n', filestr)
            return filestr
        CODE['HTML'] = HTML_code
        
        # how to typeset lists and their items in HTML:
        LIST['HTML'] = {
            'itemize':
            {'begin': '\n<ul>\n', 'item': '<li>', 'end': '</ul>\n\n'},
            'enumerate':
            {'begin': '\n<ol>\n', 'item': '<li>', 'end': '</ol>\n\n'},
            'description':
            {'begin': '\n<dl>\n', 'item': '<dt>%s<dd>', 'end': '</dl>\n\n'},
            }
        
        # how to type set description lists for function arguments, return
        # values, and module/class variables:
        ARGLIST['HTML'] = {
            'parameter': '<b>argument</b>',
            'keyword': '<b>keyword argument</b>',
            'return': '<b>return value(s)</b>',
            'instance variable': '<b>instance variable</b>',
            'class variable': '<b>class variable</b>',
            'module variable': '<b>module variable</b>',
            }
        
        # document start:
        INTRO['HTML'] = """
        <html>
        <body bgcolor="white">
        """
        # document ending:
        OUTRO['HTML'] = """
        </body>
        </html>
        """



Typesetting of Function Arguments, Return Values, and Variables
---------------------------------------------------------------

As part of comments (or doc strings) in computer code one often wishes
to explain what a function takes of arguments and what the return
values are. Similarly, it is desired to document class, instance, and
module variables.  Such arguments/variables can be typeset as
description lists of the form listed below and *placed at the end of
the doc string*. Note that argument, keyword argument, return,
instance variable, class variable, and module variable are the
only legal keywords (descriptions) for the description list in this
context.  If the output format is Epytext (Epydoc), such lists of
arguments and variables are nicely formatted using *fields* in Epytext
(this formatting requires that the list of variables appear at the end
of the doc string)::

            - argument x: x value (float),
              which must be a positive number.
            - keyword argument tolerance: tolerance (float) for stopping
              the iterations.
            - return: the root of the equation (float), if found, otherwise None.
            - instance variable eta: surface elevation (array).
            - class variable items: the total number of MyClass objects (int).
            - module variable debug: True: debug mode is on; False: no debugging 
              (bool variable).


The result depends on the output format. Epytext has special constructs
for such lists, while in the other formats we simply typeset the variable
in verbatim and keep the keywords as is.

    module variable x: 
      x value (float),
      which must be a positive number.
    module variable tolerance: 
      tolerance (float) for stopping
      the iterations.
'''
__author__ = 'Hans Petter Langtangen <hpl@simula.no>'
__version__ = 0.6

import re, os, sys, shutil, commands

from common import *
import html, latex, rst, st, epytext, plaintext
for module in html, latex, rst, st, epytext, plaintext:
    module.define(FILENAME_EXTENSION,
                  BLANKLINE,
                  INLINE_TAGS_SUBST,
                  CODE,
                  LIST,
                  ARGLIST,
                  TABLE,
                  FIGURE_EXT,
                  INTRO,
                  OUTRO)

#----------------------------------------------------------------------------
# Translators: (no, do not include! use import! - as shown above)
# include "common.py"
# include "html.py"
# include "rst.py"
# include "st.py"
# include "plaintext.py"
# include "latex.py"
#----------------------------------------------------------------------------


#debug_flag = True
debug_flag = False
if debug_flag:
    _log = open('log.log','w')
    
def debug(out):
    if debug_flag:
        #print out
        global _log
        _log = open('log.log','a')
        _log.write(out + '\n')
        _log.close()

def insert_code_from_file(filestr):
    lines = filestr.splitlines()
    inside_verbatim = False
    for i in range(len(lines)):
        line = lines[i]
        debug('Read ' + line)
        line = line.lstrip()
        if line.startswith('!bc'):
            inside_verbatim = True
        if line.startswith('!ec'):
            inside_verbatim = False
        if inside_verbatim:
            continue
            
        if line.startswith('@@@CODE'):
            debug('Found verbatim copy (line %d): %s' % (i+1, line))
            try:
                filename = line.split()[1]
            except IndexError:
                raise SyntaxError, \
                      'Syntax error: missing filename in line\n  %s' % line
            f = open(filename, 'r')
            index = line.find('fromto:')
            if index == -1:
                # no from/to regex, read the whole file:
                code = f.read()
                debug('copy the file "%s" into a verbatim block\n' % filename)
               
            else:
                patterns = line[index+7:]
                try:
                    from_, to_ = patterns.split('@')
                except:
                    raise SyntaxError, \
                    'Syntax error: missing @ in regex in line\n  %s' % line
                cfrom = re.compile(from_)
                cto = re.compile(to_)
                codelines = []
                copy = False
                for codeline in f:
                    m = cfrom.search(codeline)
                    if m:
                        copy = True
                    m = cto.search(codeline)
                    if m:
                        copy = False
                        # now the to line is not included
                    if copy:
                        debug('copy from "%s" the line\n%s' % \
                              (filename, codeline))
                        
                        codelines.append(codeline)
                code = ''.join(codelines)

            code = "!bc\n%s\n!ec\n" % code
            lines[i] = code
    
    filestr = '\n'.join(lines)
    return filestr

                
def parse_keyword(keyword, format):
    """
    Parse a keyword for a description list when the keyword may
    represent information about an argument to a function or a
    variable in a program::

      - argument x: x coordinate (float).
      - keyword argument tolerance: error tolerance (float).
      - return: corresponding y coordinate (float).
    """
    keyword = keyword.strip()
    if keyword[-1] == ':':      # strip off trailing colon
        keyword = keyword[:-1]

    typical_words = ('argument', 'keyword argument', 'return', 'variable')
    parse = False
    for w in typical_words:
        if w in keyword:    # part of function argument++ explanation?
            parse = True
            break
    if not parse:
        # no need to parse for variable type and name
        if format == 'epytext':
            # epytext does not have description lists, add a "bullet" -
            keyword = '- ' + keyword
        return keyword

    # parse:
    if 'return' in keyword:
        type = 'return'
        varname = None
        type = ARGLIST[format][type]  # formatting of keyword type
        return type
    else:
        words = keyword.split()
        varname = words[-1]
        type = ' '.join(words[:-1])
        if type == 'argument':
            type = 'parameter'
        elif type == 'keyword argument':
            type = 'keyword'
        elif type == 'instance variable':
            type = 'instance variable'
        elif type == 'class variable':
            type = 'class variable'
        elif type == 'module variable':
            type = 'module variable'
        else:
            return keyword # probably not a list of variable explanations
        # construct "type varname" string, where varname is typeset in
        # inline verbatim:
        pattern = r'(?P<begin>^)(?P<subst>%s)(?P<end>$)' % varname
        #varname = re.sub(pattern, INLINE_TAGS_SUBST[format]['verbatim'], varname)
        keyword = ARGLIST[format][type] + ' ' + varname
        return keyword


def typeset_tables(filestr, format):
    """
    | column | column | column |
    """
    from StringIO import StringIO
    result = StringIO()
    table = []
    inside_table = False
    for line in filestr.splitlines():
        if line.lstrip().startswith('|'):
            if not inside_table:
                inside_table = True
            columns = line.strip().split('|')
            # remove empty columns and extra white space:
            columns = [c.strip() for c in columns if c]
            table.append(columns)
        else:
            if inside_table:
                # not a table line anymore, but we were just inside a table
                inside_table = False
                result.write(TABLE[format](table))   # typeset table
            result.write(line)
    return result.getvalue()
            
def typeset_lists(filestr, format, debug_info=[]):
    """
    Go through filestr and parse all lists and typeset them correctly.
    This function must be called after all (verbatim) code and tex blocks
    have been removed from the file.
    """
    debug('*** List typesetting phase ***')
    from StringIO import StringIO
    result = StringIO()
    lastindent = 0
    lists = []
    inside_description_environment = False
    lines = filestr.splitlines()
    lastline = lines[0]
    # for debugging only:
    _code_block_no = 0; _tex_block_no = 0    

    for line in lines:
        
        debug('\n------------------------\nsource line=[%s]' % line)
        # do a syntax check:
        for tag in INLINE_TAGS_BUGS:
            bug = INLINE_TAGS_BUGS[tag]
            if bug:
                m = re.search(bug[0], line)
                if m:
                    print '>>> Syntax ERROR? "%s"\n    %s!' % \
                          (m.group(0), bug[1])
                
        if not line or line.isspace():  # blank line?
            result.write(BLANKLINE[format])
            debug('  > This is a blank line')
            lastline = line
            continue
        if line.startswith('#'):
            if line.startswith('#!!CODE') and len(debug_info) >= 1:
                result.write(line + '\n')
                debug('  > Here is a code block:\n%s\n--------' % \
                      debug_info[0][_code_block_no])
                _code_block_no += 1
            elif line.startswith('#!!TEX') and len(debug_info) >= 2:
                result.write(line + '\n')
                debug('  > Here is a latex block:\n%s\n--------' % \
                      debug_info[1][_tex_block_no])
                _tex_block_no += 1
            else:
                debug('  > This is just a comment line')
                # the comment can be propagated to some formats
                # (rst, LaTeX, HTML):
                line = line[1:]  # strip off initial #
                if 'comment' in INLINE_TAGS_SUBST[format]:
                    s = INLINE_TAGS_SUBST[format]['comment'] % line.strip()
                    result.write(s + '\n')

            lastline = line
            continue
            
        # structure of a line:
        linescan = re.compile(
            r"(?P<indent> *(?P<listtype>[*o-] )? *)" +
            r"(?P<keyword>[^:]+?:)?(?P<text>.*)\s?")

        m = linescan.match(line)
        indent = len(m.group('indent'))
        listtype = m.group('listtype')
        if listtype:
            listtype = listtype.strip()
            listtype = LIST_SYMBOL[listtype]
        keyword = m.group('keyword')
        text = m.group('text')
        debug('  > indent=%d (previous indent=%d)' % (indent, lastindent))

        # new (sub)section makes end of any indent (we could demand
        # (sub)sections to start in column 1, but we have later relaxed
        # such a requirement; it is easier to just set indent=0 here):
        if line.lstrip().startswith('___'):
            indent = 0


        if indent > lastindent and listtype:
            debug('  > This is a new list of type "%s"' % listtype)
            # begin a new list or sublist:
            lists.append({'listtype': listtype, 'indent': indent})
            result.write(LIST[format][listtype]['begin'])
            lastindent = indent
            if listtype == 'enumerate':
                enumerate_counter = 0

        if indent < lastindent:
            # end a list or sublist, nest back all list
            # environments on the lists stack:
            while lists and lists[-1]['indent'] > indent:
                 debug('  > This is the end of a %s list' % \
                       lists[-1]['listtype'])
                 result.write(LIST[format][lists[-1]['listtype']]['end'])
                 del lists[-1]
            lastindent = indent

        if indent == lastindent:
            debug('  > This line belongs to the previous block since it has '\
                  'the same indent (%d blanks)' % indent)

        if listtype:
            # need blank line between items and last line was not blank?
            if not (lastline.isspace() or not lastline):
                result.write(LIST[format]['separator'])

            # first write the list item identifier:
            itemformat = LIST[format][listtype]['item']
            item = itemformat
            if listtype == 'enumerate':
                debug('  > This is an item in an enumerate list')
                enumerate_counter += 1
                if '%d' in itemformat:
                    item = itemformat % enumerate_counter
                # indent here counts with '3. ':
                result.write(' '*(indent - 2 - enumerate_counter//10 - 1))  
                result.write(item + ' ')
            elif listtype == 'description':
                if '%s' in itemformat:
                    if keyword:
                        keyword = parse_keyword(keyword, format) + ':'
                        item = itemformat % keyword + ' '
                        debug('  > This is an item in a description list '\
                              'with keyword "%s"' % keyword)
                        keyword = '' # to avoid adding keyword up in
                        # below (ugly hack, but easy linescan parsing...)
                result.write(' '*(indent-2))  # indent here counts with '* '
                result.write(item)
                if not (text.isspace() or not text):
                    result.write('\n' + ' '*(indent-1))
            else:
                debug('  > This is an item in a bullet list')
                result.write(' '*(indent-2))  # indent here counts with '* '
                result.write(item + ' ')

        else:
            debug('  > This line is not part of a list environment...')
            # should check emph, verbatim, etc., syntax check and common errors
            result.write(' '*indent)      # ordinary line

        # this is not a list line and therefore we must
        # add keyword + text because these two items make up the
        # line if a : present
        if keyword:
            text = keyword + text
        debug('text=[%s]' % text)
        result.write(text + '\n')
        lastindent = indent
        lastline = line

    # end lists if any are left:
    while lists:
        debug('  > This is the end of a %s list' % lists[-1]['listtype'])
        result.write(LIST[format][lists[-1]['listtype']]['end'])
        del lists[-1]
    
    return result.getvalue()


def handle_figures(filestr, format):
    if not format in FIGURE_EXT:
        # no special handling of figures:
        return filestr
    
    pattern = INLINE_TAGS['figure']
    c = re.compile(pattern, re.MULTILINE)

    # first check if the figure files are of right type:
    files = [filename for filename, options, caption in c.findall(filestr)]
    if type(FIGURE_EXT[format]) is str:
        extensions = [FIGURE_EXT[format]]
    else:
        extensions = FIGURE_EXT[format]  # is list
    for f in files:
        basepath, ext = os.path.splitext(f)
        if not ext in extensions:
            print 'Figure', f, 'must have extension(s)', extensions
            # use convert from ImageMagick:
            for e in extensions:
                newfile = basepath + e
                if not os.path.isfile(newfile):
                    # ext might be empty, in that case we cannot convert
                    # anything:
                    if ext:
                        failure = os.system('convert %s %s' % (f, newfile))
                        if not failure:
                            print '....ok, converted %s to %s' % (f, newfile)
                            filestr = filestr.replace(f, newfile)
                            break  # jump out of inner e loop
                else:  # right file exists:
                    print '....ok, ', newfile, 'exists'
                    filestr = filestr.replace(f, newfile)
                    break
                

    # replace FIGURE... by format specific syntax:
    try:
        replacement = INLINE_TAGS_SUBST[format]['figure']
        filestr = c.sub(replacement, filestr)
    except KeyError:
        pass
    return filestr
    
def inline_tag_subst(filestr, format):
    debug('\n*** Inline tags substitution phase ***')
    ordered_tags = (
        'title', 'author', 'date', #'figure',
        # important to do section, subsection, etc. BEFORE paragraph and bold:
        'section', 'subsection', 'subsubsection',
        'emphasize', 'math2', 'math', 'bold', 'verbatim',
        'label', 'reference',
        'paragraph',  # after bold and emphasize
        # must check for plainURL before plainURL:
        'plainURL',
        # encountered unresolved problems with URLs without <link>,
        # works fine for LaTeX and HTML, but not for rst, plain
        # (unresolved problem - function caveats detects this)
        'linkURL', 
        )
    for tag in ordered_tags:
        debug('Working with tag "%s"' % tag)
        tag_pattern = INLINE_TAGS[tag]
        c = re.compile(tag_pattern, re.MULTILINE)
        try:
            replacement = INLINE_TAGS_SUBST[format][tag]
        except KeyError:
            continue  # just ignore missing tags in current format
        if replacement is None:
            continue  # no substitution
        
        if isinstance(replacement, basestring):
            occurences = len(c.findall(filestr))  # just for debug output
            filestr = c.sub(replacement, filestr)
        elif callable(replacement):
            # treat line by line because replacement string depends
            # on the match object for each occurence
            # (this is mainly for headlines in rst format)
            lines = filestr.splitlines()
            occurences = 0
            for i in range(len(lines)):
                m = re.search(tag_pattern, lines[i])
                if m:
                    replacement_str = replacement(m)
                    lines[i] = re.sub(tag_pattern, replacement_str, lines[i])
                    occurences += 1
            filestr = '\n'.join(lines)

        else:
            raise ValueError, 'replacement is of type %s' % type(replacement)
        debug('\n**** after %d "%s" substitution ***\n%s\n%s\n\n' % \
              (occurences, tag, filestr, ':'*60))
    return filestr
        

def caveats(format, filestr):
    # bug 1: plain URLs without links in rst, plain, st format
    if format in ('rst', 'st', 'plain'):
        m = re.search(INLINE_TAGS['plainURL'], filestr)
        if m:
            print 'WARNING: URLs without links, e.g., %s\n' \
           'may cause formatting errors in format "%s" (this is a bug)\n' \
           'use links instead, as in http://some.where.net<somewhere>' % \
           (m.group('url'), format)
            
    
def process(in_filename, format, out_filename):
    print '\n2nd step: run doconce2format on preprocessed file', in_filename
    f = open(in_filename, 'r')
    filestr = f.read()
    f.close()

    # 0. step: check if ^#?TITLE: is present, and if so, header and footer
    # are to be included (later below):
    if re.search(r'^#?TITLE:', filestr, re.MULTILINE):
        has_title = True
    else:
        has_title = False
        
    # 1. step: insert verbatim code from other (source code) files:
    filestr = insert_code_from_file(filestr)
    debug('%s\n**** after inserting @@@CODE:\n\n%s\n\n' % \
          ('*'*80, filestr))

    # 2. step: remove all verbatim and math blocks
    
    filestr, code_blocks, tex_blocks = remove_code_and_tex(filestr)
    # for HTML we should make replacements of < ... > in code_blocks:
    if format == 'HTML':  # fix
        from urllib import quote
        for i in range(len(code_blocks)):
            code_blocks[i] = re.sub(r'(<)([^>]*?)(>)',
                                    '&lt;\g<2>&gt;', code_blocks[i])
        
    debug('%s\n**** after removal of code/tex blocks:\n\n%s\n\n' % \
          ('*'*80, filestr))
    filestr = typeset_lists(filestr, format,
                            debug_info=[code_blocks, tex_blocks])
    debug('%s\n**** after typesetting of list:\n\n%s\n\n' % \
          ('*'*80, filestr))

    caveats(format, filestr)  # warnings

    filestr = handle_figures(filestr, format)
    
    # do substitutions:
    filestr = inline_tag_subst(filestr, format)

    """
    if format == 'LaTeX':  # fix of a seldom problem
        # if the Doconce text contains \ref{...} text it will be
        # transformed too \\ref{...} (Doconce should of course not use
        # \ref{..} but ref{...} so this is seldom a problem)
        if re.search(r'\\ref\{', filestr):
            filestr = re.sub(r'\\ref\{', r'ref{', filestr)  # tricky \, still wrong...experiment more...
        # could do \\label too
    """

    debug('%s\n**** after substituting:\n\n%s\n\n' % ('*'*80, filestr))

    filestr = insert_code_and_tex(filestr, code_blocks, tex_blocks, format)
    
    # substitute code and tex environments:
    filestr = CODE[format](filestr, format)
    debug('%s\n**** after inserting tex/code blocks:\n\n%s\n\n' % \
          ('*'*80, filestr))

    if has_title:
        if format in INTRO:
            filestr = INTRO[format] + filestr
        if format in OUTRO:
            filestr = filestr + OUTRO[format]
        
    f = open(out_filename, 'w')
    f.write(filestr)
    f.close()

def preprocess(filename, format, preprocess_options=''):
    """
    Run the preprocess script on filename and return the name
    of the resulting file.
    """
    resultfile = 'tmp.do.txt'
    print '1st step: run preprocessor on', filename
    cmd = 'preprocess -DFORMAT=%s %s %s > %s' % \
          (format, preprocess_options, filename, resultfile)
    print '>>>>', cmd
    failure, outtext = commands.getstatusoutput(cmd)
    if failure:
        print 'Could not run preprocess:\n%s' % cmd
        sys.exit(1)
    return resultfile

def main():
    try:
        format = sys.argv[1]
        filename = sys.argv[2]
    except IndexError:
        print 'Usage: %s format filename [preprocess options]\n'\
              '(formats are HTML LaTeX rst plain epytext)' \
              % sys.argv[0]
        sys.exit(1)
    debug('\n\n>>>>>>>>>>>>>>>>> %s >>>>>>>>>>>>>>>>>\n\n' % format)
    if filename[-7:] != '.do.txt':
        print 'Wrong extension of %s, must be ".do.txt"' % filename
        sys.exit(1)
    basename = filename[:-7]
    out_filename = basename + FILENAME_EXTENSION[format]
    print '\n----- doconce2format %s %s' % (format, filename)
    filename_preprocessed = preprocess(filename, format,
                                       ' '.join(sys.argv[3:]))
    process(filename_preprocessed, format, out_filename)
    os.remove(filename_preprocessed)  # clean up
    print '----- end of doconce2format: %s filtered to %s\n' % \
          (filename, out_filename)
    
if __name__ == '__main__':
    main()
